"""
The :mod:`solartime` module provides the means to calculate dawn, sunrise, solar noon, sunset,
and dusk, at a specific latitude/longitude.

The module provides one main class, :class:`SolarTime`.

:class:`SolarTime`
    * Calculates events in the UTC timezone.

Example usage ::

    >>> from datetime import date
    >>> from pytz import timezone
    >>> from solartime import SolarTime
    >>> today = date(2014, 1, 19)
    >>> localtz = timezone('US/Eastern')
    >>> lat, lon = 38.0, -79.0
    >>> sun = SolarTime()
    >>> schedule = sun.sun_utc(today, lat, lon)
    >>> sunset = schedule['sunset'].astimezone(localtz)
    >>> print(sunset)
    2014-01-19 17:24:43-05:00

"""

import datetime
from math import cos, sin, tan, acos, asin, floor, radians, degrees

try:
    import pytz
except ImportError:
    raise ImportError('The solartime module requires the pytz module to be available.')

try:
    basestring  # Python 2/3 compatibility
except NameError:
    basestring = str


__all__ = ['SolarTime', 'SolarError']

__version__ = '0.1b1'
__license__ = 'Apache 2.0'
__author__  = 'Simon Kennedy <code@sffjunkie.co.uk>, David Riggs <driggs@myotisoft.com>'


class SolarError(Exception):
    pass


class SolarTime(object):

    def __init__(self, solar_depression=6):
        """Create a SolarTime calculator.

        :param solar_depression:  Number of degrees the sun must be below the horizon for dawn/dusk calculation
        :type number or str:  Either number of degrees, or one of 'civil', 'nautical', 'astronomical'
        """

        self._depression = 6
        self.solar_depression = solar_depression  # Set default depression in degrees

    @property
    def solar_depression(self):
        """The number of degrees the sun must be below the horizon for the
        dawn/dusk calc.

        Can either be set as a number of degrees below the horizon or as
        one of the following strings

        ============= =======
        String        Degrees
        ============= =======
        civil            6.0
        nautical        12.0
        astronomical    18.0
        ============= =======
        """
        return self._depression

    @solar_depression.setter
    def solar_depression(self, depression):
        if isinstance(depression, basestring):
            try:
                self._depression = {
                    'civil': 6,
                    'nautical': 12,
                    'astronomical': 18}[depression]
            except:
                raise KeyError("solar_depression must be either a number or one of 'civil', 'nautical' or 'astronomical'")
        else:
            self._depression = float(depression)

    def sun_utc(self, date, latitude, longitude):
        """Calculate all the info for the sun at once.

        :param date:       Date to calculate for.
        :type date:        :class:`datetime.date`
        :param latitude:   Latitude - Northern latitudes should be positive
        :type latitude:    float
        :param longitude:  Longitude - Eastern longitudes should be positive
        :type longitude:   float

        :rtype:
            Dictionary with keys ``dawn``, ``sunrise``, ``noon``, ``sunset`` and ``dusk``
        """

        return {
            'dawn':    self.dawn_utc(date, latitude, longitude),
            'sunrise': self.sunrise_utc(date, latitude, longitude),
            'noon':    self.solar_noon_utc(date, longitude),
            'sunset':  self.sunset_utc(date, latitude, longitude),
            'dusk':    self.dusk_utc(date, latitude, longitude)
        }

    def dawn_utc(self, date, latitude, longitude):
        """Calculate dawn time in the UTC timezone.

        :param date:       Date to calculate for.
        :type date:        datetime.date
        :param latitude:   Latitude - Northern latitudes should be positive
        :type latitude:    float
        :param longitude:  Longitude - Eastern longitudes should be positive
        :type longitude:   float

        :rtype: date/time in UTC timezone
        """

        try:
            return self._calc_time(date, latitude, longitude, self._depression)
        except:
            raise SolarError('Sun remains below the horizon on this day, at this location.')

    def sunrise_utc(self, date, latitude, longitude):
        """Calculate sunrise time in the UTC timezone.

        :param date:       Date to calculate for.
        :type date:        datetime.date
        :param latitude:   Latitude - Northern latitudes should be positive
        :type latitude:    float
        :param longitude:  Longitude - Eastern longitudes should be positive
        :type longitude:   float

        :rtype: date/time in UTC timezone
        """

        try:
            return self._calc_time(date, latitude, longitude, 0.833)
        except:
            raise SolarError('Sun remains below the horizon on this day, at this location.')

    def solar_noon_utc(self, date, longitude):
        """Calculate solar noon time in the UTC timezone.

        :param date:       Date to calculate for.
        :type date:        datetime.date
        :param latitude:   Latitude - Northern latitudes should be positive
        :type latitude:    float
        :param longitude:  Longitude - Eastern longitudes should be positive
        :type longitude:   float

        :rtype: date/time in UTC timezone
        """

        julianday = self._julianday(date)

        newt = self._jday_to_jcentury(julianday + 0.5 + -longitude / 360.0)

        eqtime = self._eq_of_time(newt)
        timeUTC = 720.0 + (-longitude * 4.0) - eqtime

        timeUTC /= 60.0
        hour = int(timeUTC)
        minute = int((timeUTC - hour) * 60)
        second = int((((timeUTC - hour) * 60) - minute) * 60)

        if second > 59:
            second -= 60
            minute += 1
        elif second < 0:
            second += 60
            minute -= 1

        if minute > 59:
            minute -= 60
            hour += 1
        elif minute < 0:
            minute += 60
            hour -= 1

        if hour > 23:
            hour -= 24
            date += datetime.timedelta(days=1)
        elif hour < 0:
            hour += 24
            date -= datetime.timedelta(days=1)

        return datetime.datetime(date.year, date.month, date.day, hour, minute, second, tzinfo=pytz.utc)

    def sunset_utc(self, date, latitude, longitude):
        """Calculate sunset time in the UTC timezone.

        :param date:       Date to calculate for.
        :type date:        datetime.date
        :param latitude:   Latitude - Northern latitudes should be positive
        :type latitude:    float
        :param longitude:  Longitude - Eastern longitudes should be positive
        :type longitude:   float

        :rtype: date/time in UTC timezone
        """

        try:
            return self._calc_time(date, latitude, longitude, -0.833)
        except:
            raise SolarError('Sun remains below the horizon on this day, at this location.')

    def dusk_utc(self, date, latitude, longitude):
        """Calculate dusk time in the UTC timezone.

        :param date:       Date to calculate for.
        :type date:        datetime.date
        :param latitude:   Latitude - Northern latitudes should be positive
        :type latitude:    float
        :param longitude:  Longitude - Eastern longitudes should be positive
        :type longitude:   float

        :rtype: date/time in UTC timezone
        """

        try:
            return self._calc_time(date, latitude, longitude, - self._depression)
        except:
            raise SolarError('Sun remains below the horizon on this day, at this location.')

    def _julianday(self, date, timezone=None):
        day = date.day
        month = date.month
        year = date.year

        if timezone is not None:
            offset = timezone.localize(datetime.datetime(year, month, day)).utcoffset()
            offset = offset.total_seconds() / 1440.0
            day += offset + 0.5

        if month <= 2:
            year -= 1
            month += 12

        A = floor(year / 100.0)
        B = 2 - A + floor(A / 4.0)

        jd = floor(365.25 * (year + 4716)) + floor(30.6001 * (month + 1)) + day - 1524.5
        if jd > 2299160.4999999:
            jd += B

        return jd

    def _jday_to_jcentury(self, julianday):
        return (julianday - 2451545.0) / 36525.0

    def _jcentury_to_jday(self, juliancentury):
        return (juliancentury * 36525.0) + 2451545.0

    def _mean_obliquity_of_ecliptic(self, juliancentury):
        seconds = 21.448 - juliancentury * (46.815 + juliancentury * (0.00059 - juliancentury * 0.001813))
        return 23.0 + (26.0 + (seconds / 60.0)) / 60.0

    def _obliquity_correction(self, juliancentury):
        e0 = self._mean_obliquity_of_ecliptic(juliancentury)
        omega = 125.04 - 1934.136 * juliancentury
        return e0 + 0.00256 * cos(radians(omega))

    def _geom_mean_long_sun(self, juliancentury):
        l0 = 280.46646 + juliancentury * (36000.76983 + 0.0003032 * juliancentury)
        return l0 % 360.0

    def _eccentrilocation_earth_orbit(self, juliancentury):
        return 0.016708634 - juliancentury * (0.000042037 + 0.0000001267 * juliancentury)

    def _geom_mean_anomaly_sun(self, juliancentury):
        return 357.52911 + juliancentury * (35999.05029 - 0.0001537 * juliancentury)

    def _eq_of_time(self, juliancentury):
        epsilon = self._obliquity_correction(juliancentury)
        l0 = self._geom_mean_long_sun(juliancentury)
        e = self._eccentrilocation_earth_orbit(juliancentury)
        m = self._geom_mean_anomaly_sun(juliancentury)

        y = tan(radians(epsilon) / 2.0)
        y = y * y

        sin2l0 = sin(2.0 * radians(l0))
        sinm = sin(radians(m))
        cos2l0 = cos(2.0 * radians(l0))
        sin4l0 = sin(4.0 * radians(l0))
        sin2m = sin(2.0 * radians(m))

        Etime = y * sin2l0 - 2.0 * e * sinm + 4.0 * e * y * sinm * cos2l0 - 0.5 * y * y * sin4l0 - 1.25 * e * e * sin2m
        return degrees(Etime) * 4.0

    def _sun_eq_of_center(self, juliancentury):
        m = self._geom_mean_anomaly_sun(juliancentury)

        mrad = radians(m)
        sinm = sin(mrad)
        sin2m = sin(mrad + mrad)
        sin3m = sin(mrad + mrad + mrad)

        c = sinm * (1.914602 - juliancentury * (0.004817 + 0.000014 * juliancentury)) + sin2m * (0.019993 - 0.000101 * juliancentury) + sin3m * 0.000289
        return c

    def _sun_true_long(self, juliancentury):
        l0 = self._geom_mean_long_sun(juliancentury)
        c = self._sun_eq_of_center(juliancentury)
        return l0 + c

    def _sun_apparent_long(self, juliancentury):
        O = self._sun_true_long(juliancentury)

        omega = 125.04 - 1934.136 * juliancentury
        return O - 0.00569 - 0.00478 * sin(radians(omega))

    def _sun_declination(self, juliancentury):
        e = self._obliquity_correction(juliancentury)
        lambd = self._sun_apparent_long(juliancentury)

        sint = sin(radians(e)) * sin(radians(lambd))
        return degrees(asin(sint))

    def _hour_angle(self, latitude, solar_dec, solar_depression):
        latRad = radians(latitude)
        sdRad = radians(solar_dec)

        HA = (acos(cos(radians(90 + solar_depression)) / (cos(latRad) * cos(sdRad)) - tan(latRad) * tan(sdRad)))
        return HA

    def _calc_time(self, date, latitude, longitude, depression):
        julianday = self._julianday(date)

        if latitude > 89.8:
            latitude = 89.8

        if latitude < -89.8:
            latitude = -89.8

        t = self._jday_to_jcentury(julianday)
        eqtime = self._eq_of_time(t)
        solarDec = self._sun_declination(t)

        hourangle = -self._hour_angle(latitude, solarDec, 0.833)

        delta = -longitude - degrees(hourangle)
        timeDiff = 4.0 * delta
        timeUTC = 720.0 + timeDiff - eqtime

        newt = self._jday_to_jcentury(self._jcentury_to_jday(t) + timeUTC / 1440.0)
        eqtime = self._eq_of_time(newt)
        solarDec = self._sun_declination(newt)

        if depression < 0:
            depression = abs(depression)
            hourangle = -self._hour_angle(latitude, solarDec, depression)
        else:
            hourangle = self._hour_angle(latitude, solarDec, depression)

        delta = -longitude - degrees(hourangle)
        timeDiff = 4 * delta
        timeUTC = 720 + timeDiff - eqtime

        timeUTC /= 60.0
        hour = int(timeUTC)
        minute = int((timeUTC - hour) * 60)
        second = int((((timeUTC - hour) * 60) - minute) * 60)

        if second > 59:
            second -= 60
            minute += 1
        elif second < 0:
            second += 60
            minute -= 1

        if minute > 59:
            minute -= 60
            hour += 1
        elif minute < 0:
            minute += 60
            hour -= 1

        if hour > 23:
            hour -= 24
            date += datetime.timedelta(days=1)
        elif hour < 0:
            hour += 24
            date -= datetime.timedelta(days=1)

        return datetime.datetime(date.year, date.month, date.day, hour, minute, second, tzinfo=pytz.utc)



# Keep-Sabbath Copyright (c) 2019-2023 Noah Rahm.

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.

#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.

#     3. Neither the name of the copyright holder nor the names of its contributors 
#     may be used to endorse or promote products derived from this software without 
#     specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



"""
Django Keep-Sabbath is for webmasters with e-commerce shops, blogs, etc. who would like to keep the Biblical 7th-day Sabbath and other Holy days of the Bible Holy. 

The app allows you to easily setup a system for a Django-powered website that will automatically redirect visitors to a single page (explaining about the Sabbath, redirect, etc.) if it is the Sabbath day or another Holy day.

The app uses internal, offline calculations (via the SolarTime module) and the current date/time of the server to figure out if it is the Sabbath day (thus, if the user should be redirected), so you do not have to do it manually. 

However, there is an option to manually override it as well, which is useful for other Biblical Holy days.

The timing is based on the understanding that the Sabbath day starts on 6th day ("friday") evening and ends on 7th day ("saturday") evening, according to the Bible.

"If thou turn away thy foot from the sabbath, [from] doing thy pleasure on My holy day; and call the sabbath a delight, the holy of Yahweh, honourable; and shalt honour him, not doing thine own ways, nor finding thine own pleasure, nor speaking [thine own] words:
Then shalt thou delight thyself in Yahweh; and I will cause thee to ride upon the high places of the earth, and feed thee with the heritage of Jacob thy father: for the mouth of Yahweh hath spoken [it]." -Isaiah 58:13-14
"""

import pytz
from datetime import date, datetime, timedelta
from functools import wraps

# Check to see if solartime is installed
try:
    from solartime import SolarTime
except:
    # Fallback on the internal one
    from .vendor.solartime import SolarTime

from django.shortcuts import render, redirect
from django.conf import settings 
from django.utils import timezone


def _sabbath_override():
    """ Sabbath override for Holy Days, etc. when it needs 
    to be overridden to be True (this does not currently work 
    if you need to override it be False). This is not to be 
    used directly, but is used by the is_sabbath function. 
    
    :returns: a boolean value.

    This should not be called directly.
    ===================================  
    """
    override_is_sabbath = False
    try:
        if settings.SABBATH_MANUAL_OVERRIDE != False:
            override_is_sabbath = True
    except:
        raise ValueError(
            'You have not set the IS_SABBATH_OVERRIDE setting in your settings.py'
            )
    return override_is_sabbath


def _get_sunset_time(sun, localtz, day, lat, lng):
    """ Get the sunset time for a paticular location.

    :returns: a datetime object.

    This should not be called directly.
    ===================================  
    """
    schedule = sun.sun_utc(day, lat, lng)
    sunset = schedule['sunset'].astimezone(localtz)
    return sunset


def _is_during_sabbath(sun, rs_sunset_day, s_sunset_day):
    """ Internal function which uses the given data to figure out
    whether it is before, during or after the Sabbath day and returns
    the proper value.

    This should not be called directly.
    ===================================  
    """
    sabbath_start = rs_sunset_day.strftime('%y %m %d %I:%M %p')
    sabbath_end = s_sunset_day.strftime('%y %m %d %I:%M %p')
    now = timezone.now().strftime('%y %m %d %I:%M %p')

    # Uncomment below to manually set a date/time for testing
    # now = datetime(2020, 6, 9, 23, 4, 0, tzinfo=pytz.utc).strftime('%y %m %d %I:%M %p')
    # print(
    #     "sabbath starts: ", sabbath_start, 
    #     "\nsabbath ends: ", sabbath_end, 
    #     "\ndate now: ", now
    #     )

    if sabbath_start < now:
        # We know the sabbath has started
        if sabbath_end < now:
            # Its after the sabbath now
            return False
        else:
            # Its still during the sabbath
            return True 

    elif sabbath_start > now:
         # It is probably before the sabbath
        return False

    else:
        # This shouldn't ever happen... :)
        return False


def _calculate_is_sabbath(weekday):
    """ Internal function to calculate whether it is 
    the Sabbath or not. Same as calling ``is_sabbath``,
    only without the override and performance optimizations.

    This should not be called directly.
    ===================================  
    """

    # 6th day (rev Sabbath, the evening 
    # when the Sabbath starts).
    if weekday == 4: 
        rev_sabbath_day = date.today() 
        sabbath_day = rev_sabbath_day + timedelta(days=1)
    
    # the 7th day (the Sabbath)
    elif weekday == 5: 
        sabbath_day = date.today() 
        rev_sabbath_day = sabbath_day - timedelta(days=1) 
        
    # print(rev_sabbath_day)
    # print(sabbath_day)

    # Use the solartime module to calculate sunsets
    sun = SolarTime()
    localtz = pytz.timezone(settings.SABBATH_LOCAL_TIMEZONE)
    lat  = settings.SABBATH_COORDS_LATITUDE
    lng = settings.SABBATH_COORDS_LONGITUDE

    rs_sunset = _get_sunset_time(sun, localtz, rev_sabbath_day, lat, lng)
    s_sunset = _get_sunset_time(sun, localtz, sabbath_day, lat, lng)

    return _is_during_sabbath(sun, rs_sunset, s_sunset)


def get_weekday():
    """ Get the current day of the week.

    :returns: an integer of 0-6, where "Monday" 
    is 0 and "Sunday" is 6 (weird, isn't?)
    """
    return date.today().weekday()


def is_sabbath():
    """ Callable function for determining whether
    the current day and time is the Sabbath or not.
    This is based on an understanding that the Sabbath
    day starts at sundown and ends at sundown the next 
    day. 

    Usage (in a django view):
        if is_sabbath() is True:
            ...
    
    :returns: a boolean value
    """
    is_sabbath_value = False

    # See if it is currently the 6th or 7th day
    weekday = get_weekday()

    # This is a performance optimization:
    # it will only calculate the sabbath times if it
    # is near the Sabbath Day (rev-shabbat) or the day-of.
    if weekday in [4, 5]:
        is_sabbath_value = _calculate_is_sabbath(weekday)
    else:
        is_sabbath_value = False
    
    # print("Redirect?: ", is_sabbath_value)

    # Check to see if there is an override
    if _sabbath_override() == True:
        return True
    else:
        return is_sabbath_value


def _redirect_if_sabbath():
    """ Django views decorator for redirecting users to the 
    Sabbath redirect page if it is the Sabbath (or if the 
    value is overridden when it is a Holy Day, etc).
    Otherwise, the view's normal page will be returned.

    Please use ``redirect_if_sabbath`` to access this function instead
    of calling this directly.

    Usage (in views.py file):

        @redirect_if_sabbath
        def my_view(request):
            ...
    """
    def decorator(func):
        @wraps(func)
        def inner(request, *args, **kwargs):
            if is_sabbath() == True:
                return redirect('keep_sabbath:sabbath-redirect')
            return func(request, *args, **kwargs)
        return inner
    return decorator

redirect_if_sabbath = _redirect_if_sabbath()



from django.shortcuts import render
from django.conf import settings 


def keep_sabbath_page_view(request):
    try:
        return render(request, settings.SABBATH_REDIRECT_TEMPLATE, {})
    except:
        raise ValueError(
            'You have not set the SABBATH_REDIRECT_TEMPLATE in your settings.py'
            )




from setuptools import setup


setup(
  name = 'keep_sabbath',   
  packages = [
    'keep_sabbath',
    'keep_sabbath.vendor',
  ],   
  version = '0.7.0',
  license='BSD 3-Clause',   
  description = 'Django app to automatically redirect users when it is the Sabbath day, with support for overriding on other Biblical Holy days.',
  long_description=open('README.md').read(),
  long_description_content_type='text/markdown',
  author = 'Noah Rahm', 
  author_email = 'hi@noahrahm.com',     
  url = 'https://github.com/Correct-Syntax/Keep-Sabbath/', 
  keywords = ['sabbath', 'django', 'web', 'redirects'], 
  install_requires=[           
    'django>=3.2'
  ],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Operating System :: OS Independent',
    'Environment :: Web Environment',
    'Framework :: Django',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: BSD License',   
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
  ],
)

